from pathlib import Path
import re
import traceback

from click.testing import CliRunner, Result
from pytest import fixture

RE_BLOCK_START = re.compile(r'^\[[^]]*test_name="?([^"]*)"?[^]]*]$')
RE_INTERN_BLOCK = re.compile(
    r'^\| \[source, ?([a-z]+)]\n-{4,}\n((.|\n)*?)\n-{4,}$', re.MULTILINE
)
RE_TABLE_DELIM = re.compile(r'^\|={3,}$')


@fixture(scope='function')
def project_result(request, monkeypatch) -> Result:
    from syntrend.cli import generate

    name = request.node.name.split('_', 1)[1]
    runner = CliRunner(mix_stderr=False)
    test_config = Path(f'tests/assets/uc_{name}.yaml').absolute()
    with runner.isolated_filesystem():
        cli_result = runner.invoke(generate, [str(test_config)])
    if cli_result.exit_code:
        traceback.print_tb(cli_result.exc_info[2])
    return cli_result



from syntrend.generators import register, PropertyGenerator


@register
class FloatGenerator(PropertyGenerator):
    type = float
    name = 'float'
    default_config = {
        'min_offset': -500.0,
        'max_offset': 500.0,
        'num_decimals': 6,
    }
    required_modules = ['random']

    def load_kwargs(self, kwargs: dict[str, any]) -> dict[str, any]:
        kwargs['min_offset'] = float(kwargs['min_offset'])
        kwargs['max_offset'] = float(kwargs['max_offset'])
        kwargs['num_decimals'] = int(kwargs['num_decimals'])
        kwargs['range'] = kwargs['max_offset'] - kwargs['min_offset']
        return kwargs

    def validate(self):
        if self.kwargs.min_offset > self.kwargs.max_offset:
            raise ValueError('Min Offset must be less than or equal to Max Offset')

    def generate(self):
        return round(
            self.modules.random.random() * self.kwargs.range + self.kwargs.min_offset,
            self.kwargs.num_decimals,
        )



from syntrend.config import model
import syntrend.generators as generators
import syntrend.generators.string as string_gen

from functools import partial

from pytest import mark, fixture

Prop_Def = partial(model.PropertyDefinition, name='test')


@fixture(scope='function', autouse=True)
def load_generators():
    generators.load_generators()


@mark.unit
def test_load_default_generator():
    base_def = Prop_Def(type='string', **generators.GENERATORS['string'].default_config)
    prop_def = Prop_Def(type='string')
    str_gen = generators.get_generator('test', prop_def, None)
    exp_gen = string_gen.StringGenerator('test', base_def)
    assert str_gen.__class__.__name__ == exp_gen.__class__.__name__, (
        'Load Generator should return the String Generator'
    )


@generators.register
class RandomTestGenerator(generators.PropertyGenerator):
    name = 'test'

    def generate(self):
        return 1, 'test', self.config.kwargs


@mark.unit
def test_register_generator():
    prop_def = Prop_Def(type='test')
    exp_gen = RandomTestGenerator('test', prop_def)
    try:
        test_gen = generators.get_generator('test', prop_def, None)
    except KeyError:
        raise ValueError("Should have found a 'test' generator") from None
    assert test_gen.__class__ == exp_gen.__class__, (
        "Should have returned the 'test' generator"
    )



from syntrend.utils import exc
from os import linesep


def test_exception_output(monkeypatch):
    reported_exception = [0]
    reported_output = ['']

    def _exit_code(code: int):
        reported_exception[0] = code

    def _error_output(content: str):
        reported_output[0] = content

    monkeypatch.setattr(exc.sys, 'exit', _exit_code)
    monkeypatch.setattr(exc.sys.stderr, 'write', _error_output)
    err = ValueError('Failed Expression', {'a': 1, 'b': 2, 'c': 3})
    exc.process_exception(err)

    assert reported_exception[0] == 1, (
        'Reported error exit code should be 2 (Expression Error)'
    )
    assert (
        reported_output[0]
        == linesep.join(
            [
                'Error Encountered: (ValueError)',
                '  | Failed Expression',
                '  | a: 1',
                '  | b: 2',
                '  | c: 3',
            ]
        )
        + linesep
    ), 'Exception output format not meeting expected format'



from syntrend.formatters import register_formatter, Event, Collection
from syntrend.config import CONFIG

from functools import cache

cell_format = '{l_buffer}{value}{r_buffer}'
row_format = ' {cols} '


@register_formatter('table')
def table_formatter(object_name: str):
    output_options = CONFIG.objects[object_name].output
    col_sep = output_options.kwargs.get('column_separator', ' ')
    if col_sep != ' ':
        col_sep = f' {col_sep} '
    column_widths: dict[str, int] = {}

    @cache
    def table_width():
        return (
            sum(column_widths.values()) + (len(col_sep) * (len(column_widths) - 1)) + 2
        )
        # Table Width = (total of max widths) + ( (width of separator) * (# cols - 1) ) + 2 extra spaces at start/end

    def _write_rows(events: Collection, row_sep: str) -> list[str]:
        buffer = []
        for event in events:
            row = []
            for key, value in event.items():
                l_adjust = isinstance(value, (str,))
                cell_buffer = ' ' * (column_widths[key] - len(str(event[key])))
                row.append(
                    cell_format.format(
                        l_buffer='' if l_adjust else cell_buffer,
                        r_buffer='' if not l_adjust else cell_buffer,
                        value=value,
                    )
                )
            buffer.append(row_format.format(cols=col_sep.join(row)))
            if row_sep:
                buffer.append(row_sep * table_width())
        return buffer

    def __formatter(events: Collection) -> list[str]:
        buffer = []
        column_widths.clear()
        for event in events:
            widths: dict[str, int] = {}
            for key, value in event.items():
                cell_width = len(str(value))
                column_widths[key] = max(
                    len(key), column_widths.get(key, 0), cell_width
                )
                widths[key] = cell_width
        if output_options.collection:
            buffer += _write_rows(
                events=Collection(Event({key: key for key in column_widths})),
                row_sep=output_options.kwargs.get('header_separator', '='),
            )
        buffer += _write_rows(
            events=events,
            row_sep=output_options.kwargs.get('row_separator', ''),
        )
        return buffer

    return __formatter



from syntrend.config import CONFIG
from syntrend.formatters import register_formatter, Collection

ROW_FORMAT = '{row_indent}{content}{sep}'


@register_formatter('json')
def json_formatter(object_name: str):
    import json

    is_collection = CONFIG.objects[object_name].output.collection

    def __formatter(events: Collection) -> list[str]:
        buffer = ['['] if is_collection else []
        row_buffer = '  ' if is_collection else ''
        event_count = len(events)
        for idx, event in enumerate(events):
            buffer.append(
                ROW_FORMAT.format(
                    row_indent=row_buffer,
                    content=json.dumps(event['value'] if event.use_default else event),
                    sep=',' if idx < event_count - 1 and is_collection else '',
                )
            )
        if is_collection:
            buffer.append(']')
        return buffer

    return __formatter



"""Project File Schema"""

import logging
from typing import Union, Optional
import dataclasses as dc
from enum import Enum
from os import getenv
from pathlib import Path
from functools import partial

LOG = logging.getLogger(__name__)
DEFAULT_ENV_VAR_PREFIX = 'SYNTREND_'
OUTPUT_STDOUT = Path('-')
DEFAULT_FILE_FORMAT = '{name}-{id}.{format}'
USER_CONFIG_DIR = Path.home().joinpath('.config', 'syntrend')
ADD_GENERATOR_DIR = USER_CONFIG_DIR.joinpath('generators')

dataclass = partial(dc.dataclass, kw_only=True, init=False)


class NullValue:
    pass


class _NullInt(NullValue, int):
    pass


class _NullString(NullValue, str):
    pass


NULL_VAL = NullValue()
NULL_INT = _NullInt(0)
NULL_STR = _NullString('')


class DistributionTypes(Enum):
    NoDistribution = 'none'
    Linear = 'linear'
    StandardDeviation = 'std_dev'


class Progress(Enum):
    Trend = 'trend'
    Last = 'last'


def fields(obj: dc.dataclass, include_field=False) -> list[str | dc.Field]:
    return [field if include_field else field.name for field in dc.fields(obj)]


@dataclass
class Validated:
    """Base Class for Configurations"""

    source__: dict = dc.field(default_factory=dict)

    def __init__(self, **kwargs):
        """Runs Parsing methods (if declared) to parse dataclass fields and validation method. Requires
        a method to match the field name with the following signature:
          `parse_<field.name>(self, value) -> any`
        """
        kwargs.update(kwargs.pop('kwargs', {}))
        self.source__ = kwargs.copy()
        for field in fields(self, include_field=True):
            default_val = NULL_VAL
            if field.default is not dc.MISSING:
                default_val = field.default
            elif field.default_factory is not dc.MISSING:
                default_val = field.default_factory()
            setattr(self, field.name, kwargs.pop(field.name, default_val))
            if callable(method := getattr(self, f'parse_{field.name}', None)):
                setattr(self, field.name, method(getattr(self, field.name)))
        self.kwargs = kwargs
        if hasattr(self, 'parse_kwargs'):
            self.kwargs = self.parse_kwargs(kwargs)
        self.validate()

    def __repr__(self):
        _fields = [
            f'{field_name}={repr(getattr(self, field_name))}'
            for field_name in fields(self) + ['kwargs']
        ]
        return f'<{type(self).__name__}({_fields})>'

    def validate(self):
        return


def copy(obj: Validated) -> Validated:
    """Generates a duplicate of an object

    Returns:
        Instance of a `Validated` subclass
    """
    new_dict = {
        f_name: copy(f_val) if isinstance(f_val, Validated) else f_val
        for f_name, f_val in [
            (fld_name, getattr(obj, fld_name)) for fld_name in fields(obj)
        ]
        if not obj.source__ or f_name in obj.source__
    }
    return type(obj)(**(new_dict | obj.kwargs))


def update(obj: Validated, other: Validated) -> None:
    """Applies any values from one `Validated` instance into another.

    Similar to `dict.update()` but applies specifically to `Validated` instances to preserve
    class behaviour

    Args:
        obj: Instance of `Validated` subclass to update
        other: Instance of `Validated` subclass to copy values from

    Raises:
        TypeError: `other` is not a subclass of `Validated`
    """
    if not isinstance(other, Validated):
        raise TypeError(
            'Only `Validated` subclasses can be supported to update from',
            {
                'Original Object Type': type(obj).__name__,
                'Other Object Type': type(other).__name__,
            },
        )

    for field in fields(obj):
        setattr(obj, field, getattr(other, field))
    obj.kwargs.update(other.kwargs)
    obj.source__ = other.source__


def parse_int(_min: Optional[int] = None, _max: Optional[int] = None):
    """Convenience function to parse integer values for `Validated` classes

    Args:
        _min: Minimum value of the integer range
        _max: Maximum value of the integer range

    Returns:
        Callable WrappParsed integer for the field

    Raises:
        TypeError: Input Value is not a valid Integer type
        ValueError: Input Value is not within the defined range
    """

    def _parser(_, value):
        try:
            value = int(value)
        except TypeError:
            raise TypeError(
                'Value must be parsable to integer',
                {
                    'Input Value': str(value),
                    'Input Value Type': type(value).__name__,
                },
            ) from None
        if _min is not None and value < _min:
            raise ValueError(
                'Provided value is less than the minimum allowed',
                {
                    'Input Value': str(value),
                    'Minimum': _min,
                },
            )
        if _max is not None and value > _max:
            raise ValueError(
                'Provided value is greater than the maximum allowed',
                {
                    'Input Value': str(value),
                    'Maximum': _max,
                },
            )
        return value

    return _parser


@dataclass
class ModuleConfig(Validated):
    """Configuration Properties to modify/alter how the `syntrend` utility behaves"""

    max_generator_retries: int = dc.field(
        default=int(getenv(f'{DEFAULT_ENV_VAR_PREFIX}_MAX_GENERATOR_RETRIES', 20))
    )
    """Maximum number of retries a Generator can perform before failing."""
    max_historian_buffer: int = dc.field(
        default=int(getenv(f'{DEFAULT_ENV_VAR_PREFIX}_MAX_HISTORIAN_BUFFER', 20))
    )
    """Maximum values to be kept in a buffer of previous values"""
    generator_dir: str = dc.field(
        default=getenv(f'{DEFAULT_ENV_VAR_PREFIX}_GENERATOR_DIR', '')
    )
    """Source Directory of Custom Generators"""

    parse_max_generator_retries = parse_int(_min=1)
    parse_max_historian_buffer = parse_int(_min=1)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def parse_generator_dir(self, value: str) -> Path:
        if not value:
            return ADD_GENERATOR_DIR
        parsed_path = Path(value).absolute()
        if not parsed_path.is_dir():
            raise ValueError(
                'Source Generator Directory does not exist',
                {
                    'Input Path': value,
                    'Parsed Path': str(parsed_path),
                },
            )
        return parsed_path


@dataclass
class OutputConfig(Validated):
    """Configuration Properties used for Global and Object-specific outputs"""

    format: str = dc.field(default='json')
    directory: Path = dc.field(default='-')
    filename_format: str = dc.field(default='{name}_{id}.{format}')
    collection: bool = dc.field(default=False)
    count: int = dc.field(default=1)
    time_field: str = dc.field(default='')

    def parse_collection(self, value):
        return bool(value)

    def parse_directory(self, value):
        if isinstance(value, Path):
            return value
        if value == '-':
            return OUTPUT_STDOUT
        p = Path(value).absolute()
        if not p.exists():
            p.mkdir(parents=True)
        assert p.is_dir(), 'Path must be a directory'
        return p


# class ValueRange(BaseModel):
#     min: T_VALUE
#     max: T_VALUE
#     distribution: DistributionTypes = DistributionTypes.Linear
#     factor: float = Field(0.0)
#
#     def __init__(self, **kwargs):
#         if (
#                 "distribution" in kwargs
#                 and isinstance(kwargs["distribution"], str)
#                 and "factor" not in kwargs
#                 and DIST_SEP in kwargs["distribution"]
#         ):
#             parts = kwargs["distribution"].split(DIST_SEP)
#             kwargs["factor"] = parts[1]
#             kwargs["distribution"] = parts[0]
#         super().__init__(**kwargs)


@dataclass
class PropertyDistribution(Validated):
    """Distribution Definition

    Configurations to support how values will vary from its original value

    Attributes:
        type (:obj:`DistributionTypes`): The type of distribution to apply. Defaults to "none"
        std_dev (:obj:`float`, optional): The standard deviation of the distribution
        min_offset (:obj:`float`, optional): The minimum offset of the distribution
        max_offset (:obj:`float`, optional): The maximum offset of the distribution
    """

    type: DistributionTypes = DistributionTypes.NoDistribution
    std_dev: float = 0.0
    min_offset: Union[int, float] = 0
    max_offset: Union[int, float] = 1

    def parse_type(self, value):
        if isinstance(value, DistributionTypes):
            return value
        return DistributionTypes(value)

    def validate(self):
        if self.min_offset > self.max_offset:
            raise ValueError(
                'Distribution Min value must be lower than the Max value',
                {
                    'Minimum Value': self.min_offset,
                    'Maximum Value': self.max_offset,
                },
            )


@dc.dataclass
class PropertyDefinition(Validated):
    """Property Definition

    Definition of how a value is generated and any associated properties to modify its behaviour

    Attributes:
        type (:obj:`str`): Generator Type to be used for this Property Definition
        distribution (:obj:`PropertyDistribution`): Property to define how the generated values will vary using a
            :obj:`DistributionTypes`. Defaults to "none"
        expression (:obj:`str`): String Expression to define a trend, behaviour, or conditions to apply.

            See Also:
                For more information on Expressions, see `Expressions <docs/expressions>`__.
        start: Any value associated with :obj:`type` for when a previous value is expected but none available.
        items (:obj:`list`): List of items required for Generator Types needing a list of objects to choose from.
        properties (:obj:`dict[str, PropertyDefinition]`): Mapping of sub properties namely to support nested objects.
    """

    name: str
    type: str
    distribution: Union[DistributionTypes, PropertyDistribution] = dc.field(
        default=DistributionTypes.NoDistribution
    )
    conditions: list[str] = dc.field(default_factory=list)
    expression: str = dc.field(default='')
    start: any = dc.field(default=None)
    items: list[any] = dc.field(default_factory=list)
    properties: dict[str, 'PropertyDefinition'] = dc.field(default_factory=dict)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def parse_distribution(self, dist_type):
        if isinstance(dist_type, str):
            dist_type = DistributionTypes(dist_type)
        if isinstance(dist_type, DistributionTypes):
            dist_type = {'type': dist_type}
        if isinstance(dist_type, dict):
            dist_type = PropertyDistribution(**dist_type)
        return dist_type

    def parse_properties(self, props):
        return {
            prop: (
                val
                if isinstance(val, PropertyDefinition)
                else PropertyDefinition(
                    name=prop if 'name' not in val else val['name'], **val
                )
            )
            for prop, val in props.items()
        }


@dataclass
class ObjectDefinition(PropertyDefinition):
    """Object Definition

    Extended definition of :obj:`PropertyDefinition` to support root-level object behaviour

    Attributes:
        output (:obj:`OutputConfig`): Properties to define how and where results are generated
    """

    output: OutputConfig = dc.field(default_factory=OutputConfig)

    def parse_output(self, value):
        if isinstance(value, OutputConfig):
            return value
        return OutputConfig(**value)


@dataclass
class ProjectConfig(Validated):
    """Project File Configuration

    Root-Level configuration for project files.

    Attributes:
        objects (:obj:`dict[str, ObjectDefinition]`): Mapping of Object Definitions
        output (:obj:`OutputConfig`): Default properties to define how and where results are generated
        config (:obj:`ModuleConfig`): Syntrend configuration properties to modify tool behaviour
    """

    objects: dict[str, ObjectDefinition]
    output: OutputConfig = dc.field(default_factory=OutputConfig)
    config: ModuleConfig = dc.field(default_factory=ModuleConfig)

    def parse_output(self, output):
        if isinstance(output, OutputConfig):
            return output
        return OutputConfig(**output)

    def parse_config(self, config):
        if isinstance(config, ModuleConfig):
            return config
        return ModuleConfig(**config)

    def parse_objects(self, objects):
        if len(objects) == 0:
            raise ValueError('Project Config must include one object to generate', {})
        return {
            obj_name: ObjectDefinition(name=obj_name, **objects[obj_name])
            for obj_name in objects
        }



from syntrend.formatters import json, Event, Collection

from pytest import mark


@mark.unit
def test_single_string(project, monkeypatch):
    project(json, {'type': 'string'})
    formatter = json.json_formatter('test')
    output = formatter(Collection(Event('generated_string')))
    assert output[0] == '"generated_string"', 'Should generate a quoted string'


@mark.unit
def test_single_number(project, monkeypatch):
    project(json, {'type': 'integer'})
    formatter = json.json_formatter('test')
    output = formatter(Collection(Event(10)))
    assert output[0] == '10', 'Should generate an encoded number string'


@mark.unit
def test_single_object(project, monkeypatch):
    project(json, {'type': 'object'})
    formatter = json.json_formatter('test')
    output = formatter(Collection(Event({'f1': 'string', 'f2': 10})))
    assert output[0] == '{"f1": "string", "f2": 10}', (
        'Should generate an encoded json string of multiple values'
    )


@mark.unit
def test_multiple_objects(project, monkeypatch):
    project(json, {'type': 'object'})
    formatter = json.json_formatter('test')
    output = formatter(
        Collection(
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
        )
    )
    assert output[0] == '{"f1": "string", "f2": 10}', (
        'Should generate an encoded json string of multiple values'
    )
    assert len(output) == 3, (
        'Should generate 3 lines of json output with an extra empty line'
    )


@mark.unit
def test_multiple_objects_as_collection(project, monkeypatch):
    project(json, {'type': 'object', 'output': {'collection': True}})
    formatter = json.json_formatter('test')
    output = formatter(
        Collection(
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
        )
    )
    assert output[0] == '[', 'First line should include a list encapsulation'
    assert output[1] == '  {"f1": "string", "f2": 10},', (
        'Should generate an indented json string of multiple values'
    )
    assert len(output) == 5, (
        'Should generate 3 lines of json output with 2 lines for list encapsulation, and an extra empty line'
    )



import matplotlib.pyplot as plt
from matplotlib import colors
import numpy as np
import sys
from typing import Callable
from functools import partial
from pathlib import Path

T_IMG = tuple[float, float, float]
OUT_PATH = Path.cwd().joinpath('docs', 'assets')


def smooth1d(x, window_len):
    s = np.r_[2 * x[0] - x[window_len:1:-1], x, 2 * x[-1] - x[-1:-window_len:-1]]
    w = np.hanning(window_len)
    y = np.convolve(w / w.sum(), s, mode='same')
    return y[window_len - 1 : -window_len + 1]


def gaussian_filter(
    sigma: float, alpha: float = 0.5, color: T_IMG = (0.0, 0.0, 0.0)
) -> Callable[[T_IMG, float], tuple[T_IMG, float, float]]:
    def _run_filter(image: T_IMG, dpi: float) -> tuple[T_IMG, float, float]:
        pad = int(sigma * 3 / 72.0 * dpi)
        padded_src = np.pad(image, [(pad, pad), (pad, pad), (0, 0)], 'constant')
        target_image = np.empty_like(padded_src)
        target_image[:, :, :3] = color

        scaled_sigma = sigma / 72.0 * dpi
        alpha_src = padded_src[:, :, -1] * alpha
        window_len = max(int(scaled_sigma), 3) * 2 + 1
        alpha_src = np.apply_along_axis(smooth1d, 0, alpha_src, window_len)
        # alpha_src = np.apply_along_axis(smooth1d, 1, alpha_src, int(window_len / 10))
        ratio = alpha / np.max(alpha_src)
        target_image[:, :, 3] = alpha_src * ratio

        # margin = pad * .1
        # inner_pad = int(pad - margin)
        # target_image = target_image[inner_pad:-inner_pad, inner_pad:-inner_pad]
        return target_image, -pad, -pad

    return _run_filter


def apply_gaussian_blur(plot: plt.Axes, sigma: float, alpha: float, **kwargs):
    (lines,) = plot.get_lines()
    yy = lines.get_ydata()
    gauss = gaussian_filter(
        sigma=sigma, alpha=alpha, color=colors.to_rgb(kwargs['color'])
    )
    (line,) = plot.plot(lines.get_xdata(), yy)
    line.update_from(lines)
    line.set_agg_filter(gauss)
    line.set_rasterized(True)
    line.update(kwargs)


GRAPH_TYPE = ''
X = np.linspace(0, 29, 30)
X_TREND = np.array([5.0] * 30)
Y = {
    'none': [5.0] * 30,
    'linear': [
        3.4286597355069164,
        2.759552727927165,
        5.139106683692621,
        7.846772286339835,
        4.370221795766835,
        6.163291543687663,
        5.760275442542536,
        4.883550131468994,
        7.958263680754884,
        2.4737366106868652,
        8.50466700352673,
        5.8691653857392545,
        7.880160966544324,
        7.9982299762617295,
        7.463460637918457,
        3.9290881421875508,
        8.894258971032642,
        8.955755609952814,
        4.288907058893064,
        5.388089616412426,
        6.559485074178385,
        6.496753657232439,
        9.881486962248243,
        6.2532589849034395,
        5.245457904462363,
        7.730317837887085,
        7.8299044838888285,
        4.727513065234957,
        5.19795733463555,
        3.757641452375263,
    ],
    'std_dev': [
        6.539895504932106,
        5.864296778185984,
        6.58124836310444,
        1.5985193382969385,
        4.711970924784011,
        6.6513144302880285,
        2.9541544268680777,
        5.483990056762898,
        5.772204523194729,
        6.058900154473583,
        0.5072872916252509,
        5.512192691464881,
        7.172460636216519,
        6.672910636744286,
        7.338767339760278,
        5.735282162631125,
        7.269013928563343,
        6.961476081597453,
        4.439704547471921,
        1.6496701535925196,
        0.9593543409724301,
        1.9368649803801063,
        6.9704527608035995,
        4.686448324720903,
        6.289802586590157,
        3.3627063781552518,
        4.038932129681856,
        6.035254516582815,
        6.281155903537542,
        6.675274226713796,
    ],
}
AREA = {
    'none': lambda plot: None,
    'linear': lambda plot: partial(
        plot.fill_between, X, y1=X_TREND + 5, y2=X_TREND - 3, alpha=0.2
    ),
    'std_dev': lambda plot: partial(apply_gaussian_blur, plot, alpha=0.2, sigma=70),
}


def draw_value_diagram(plot: plt.Axes):
    values = np.array(Y[GRAPH_TYPE])
    plot.tick_params(axis='y', left=False, labelleft=False)
    plot.grid()
    (trend_line,) = plot.plot(X, X_TREND, color='blue', label='Trend Expression')
    dist_area = AREA[GRAPH_TYPE](plot)
    if dist_area:
        dist_area(label='Area of Distribution', color=trend_line.get_color())
    plot.plot(X, values, 'o', color='green', label='Generated Values')
    plot.set_ylim((-5, 15))
    plot.legend(loc='best', fancybox=True)


DIST_BIN_SIZE = 0.01
ALPHA = 2
DIST_GRAPH = {
    'none': lambda: np.array(
        [
            100 if i * DIST_BIN_SIZE == 5 else 0
            for i in range(0, int(10 / DIST_BIN_SIZE))
        ]
    ),
    'linear': lambda: np.array(
        [
            100 if 2 <= i * DIST_BIN_SIZE <= 10 else 0
            for i in range(0, int(10 / DIST_BIN_SIZE))
        ]
    ),
    'std_dev': lambda: 1.0
    / (np.sqrt(2.0 * np.pi) * 2.0)
    * np.exp(-np.power((np.arange(0.0, 10.0, DIST_BIN_SIZE) - 5.0) / 2.0, 2.0) / 2),
}


def draw_distribution_graph(plot: plt.Axes):
    plot.set_title('Dist @ 0')
    plot.tick_params(axis='y', bottom=False, labelbottom=False)
    plot.grid()
    plot.fill_betweenx(
        np.arange(0.0, 10.0, DIST_BIN_SIZE),
        DIST_GRAPH[GRAPH_TYPE](),
        0,
        color='blue',
        alpha=0.2,
    )
    plot.axhline(5, color='blue')


def draw_plot():
    fig = plt.figure(figsize=(6, 3))
    grid_spec = fig.add_gridspec(
        1, 2, width_ratios=(1, 4), left=0.1, right=0.9, bottom=0.1, top=0.9, wspace=0.05
    )
    pl_trend = fig.add_subplot(grid_spec[0, 1])
    pl_dist = fig.add_subplot(grid_spec[0, 0], sharey=pl_trend)
    draw_value_diagram(pl_trend)
    draw_distribution_graph(pl_dist)
    fig.savefig(OUT_PATH.joinpath(f'distribution_graph_{GRAPH_TYPE}.png'))
    # fig.show()


def none():
    global GRAPH_TYPE
    GRAPH_TYPE = 'none'
    draw_plot()


def linear():
    global GRAPH_TYPE
    GRAPH_TYPE = 'linear'
    draw_plot()


def std_dev():
    global GRAPH_TYPE
    GRAPH_TYPE = 'std_dev'
    draw_plot()


if __name__ == '__main__':
    print('Plot Call', sys.argv)
    cmd, GRAPH_TYPE = sys.argv
    draw_plot()



from syntrend.formatters import sql, Event, Collection

from pytest import mark


@mark.unit
def test_single_string(project, monkeypatch):
    project(sql, {'type': 'string'})
    formatter = sql.sql_formatter('test')
    output = formatter(Collection(Event('generated_string')))
    assert output[0] == 'insert into test (value) values ("generated_string")', (
        'Should generate an insert statement'
    )


@mark.unit
def test_single_number(project, monkeypatch):
    project(sql, {'type': 'integer'})
    formatter = sql.sql_formatter('test')
    output = formatter(Collection(Event(10)))
    assert output[0] == 'insert into test (value) values (10)', (
        'Should generate an insert statement'
    )


@mark.unit
def test_single_object(project, monkeypatch):
    project(sql, {'type': 'object'})
    formatter = sql.sql_formatter('test')
    output = formatter(Collection(Event({'f1': 'string', 'f2': 10})))
    assert output[0] == 'insert into test (f1, f2) values ("string", 10)', (
        'Should generate an insert statement with 2 fields'
    )


@mark.unit
def test_multiple_objects(project, monkeypatch):
    project(sql, {'type': 'object'})
    formatter = sql.sql_formatter('test')
    output = formatter(
        Collection(
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
        )
    )
    assert all(
        [line == 'insert into test (f1, f2) values ("string", 10)' for line in output]
    ), 'All lines should generate an insert statement'
    assert len(output) == 3, 'Should generate 3 lines of sql output'


@mark.unit
def test_multiple_objects_as_collection(project, monkeypatch):
    project(sql, {'type': 'object', 'output': {'collection': True}})
    formatter = sql.sql_formatter('test')
    output = formatter(
        Collection(
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
        )
    )
    assert all(
        [line == 'insert into test (f1, f2) values ("string", 10)' for line in output]
    ), 'All lines should generate an insert statement'
    assert len(output) == 3, (
        'Should generate 3 lines of sql output, no changes for collections'
    )



from syntrend.config import model

from typing import Union, Generator, Any
from pathlib import Path

from ruamel.yaml import YAML, error

yaml = YAML(typ='safe')
CONFIG_MODULES = [
    model.ProjectConfig,
    model.OutputConfig,
    model.ModuleConfig,
]
MODULE_KEYS = {}
for module in CONFIG_MODULES:
    yaml.register_class(module)
    MODULE_KEYS[module] = set(model.fields(module))


def parse_object(config_dict: dict) -> Union[dict, model.Validated]:
    dict_keys = set(config_dict)
    for _module in MODULE_KEYS:
        if not dict_keys - MODULE_KEYS[_module]:
            return _module(**config_dict)
    return config_dict


def retrieve_source(
    config_file: Union[list[dict], dict, str, Path],
) -> Generator[dict[str, Any], None, None]:
    if isinstance(config_file, list):
        for inner_dict in config_file:
            yield parse_object(inner_dict)
        return
    if isinstance(config_file, dict):
        yield parse_object(config_file)
        return

    content = config_file
    if (path_ref := Path(config_file)).exists():
        with path_ref.open('r') as file_obj:
            content = file_obj.read()

    try:
        for doc in yaml.load_all(content):
            yield doc
    except error.YAMLError as err:
        raise ValueError(
            f'Invalid content format provided for parsing - {type(err).__name__}: {err.args}',
            {'File Header': content[:20]},
        ) from None


def load_config(config_file: Union[dict, str, Path]) -> model.ProjectConfig:
    new_config = model.ProjectConfig(objects={'test': {'type': 'string'}})
    for config_obj in retrieve_source(config_file):
        if isinstance(config_obj, model.ProjectConfig):
            new_config = config_obj
        elif isinstance(config_obj, model.OutputConfig):
            model.update(new_config.output, config_obj)
        elif isinstance(config_obj, model.ModuleConfig):
            model.update(new_config.config, config_obj)
        elif isinstance(config_obj, dict):
            if 'type' in config_obj:
                config_obj = {'this': config_obj}
            new_config.objects = {
                obj_name: model.ObjectDefinition(name=obj_name, **config_obj[obj_name])
                for obj_name in config_obj
            }
        else:
            raise ValueError(
                f'Unhandled Configuration Type: {repr(config_obj)}',
                {'Config Object Type': type(config_obj).__name__},
            )
    return new_config



from syntrend.config import CONFIG
from collections import deque
import re

RE_PATH_ROOT = re.compile(r'^(?:\[(-?\d*)])?\.?(.*)$')


class Historian:
    def __init__(self):
        self.__values: deque = deque(maxlen=CONFIG.config.max_historian_buffer)

    # Historian Manager methods
    def append(self, _object):
        self.__values.appendleft(_object)

    def has_values(self) -> bool:
        return len(self.__values) > 0

    def __getitem__(self, item: int):
        return self.__values[item]

    ##########################################################
    # Convenience Operators when defaulting to current value #
    ##########################################################
    @property
    def current(self):
        return self.__values[0]

    # Comparator Methods
    def __eq__(self, other):
        return self.current == other

    def __ne__(self, other):
        return self.current != other

    def __lt__(self, other):
        return self.current < other

    def __le__(self, other):
        return self.current <= other

    def __gt__(self, other):
        return self.current > other

    def __ge__(self, other):
        return self.current >= other

    # Arithmetic Operators
    def __add__(self, other):
        return self.current + other

    def __sub__(self, other):
        return self.current - other

    def __mul__(self, other):
        return self.current * other

    # Formatting / Type Casting
    def __len__(self):
        return len(self.__values)

    def __int__(self):
        return int(self.current)

    def __float__(self):
        return float(self.current)

    def __hex__(self):
        current = self.current
        if isinstance(current, int):
            return hex(current)
        if isinstance(current, float):
            return current.hex()
        raise TypeError(
            'Value in this series cannot be converted to Hexadecimal',
            {'Value': str(current), 'Value Type': type(current).__name__},
        )

    def __format__(self, format_spec):
        return format(self.current, format_spec)

    def __str__(self):
        return f'<Historian current={str(self.current)}>'

    def __repr__(self):
        return f'<Historian current={str(self.current)}>'

    # Removing unwanted methods
    def insert(self, _, __):
        raise AttributeError("'Historian' object has no attribute 'insert'")

    def __iter__(self):
        raise AttributeError("'Historian' object has no attribute '__iter__'")

    def __next__(self):
        raise AttributeError("'Historian' object has no attribute '__next__'")

    def __contains__(self, item):
        raise AttributeError("'Historian' object has no attribute '__contains__'")



from syntrend.config import model

from pytest import mark, raises


@mark.unit
@mark.parametrize('value,_min,_max', [(1, 0, 2), (1, 1, 2), ('2', -2, 2), (-3, -5, -1)])
def test_parse_int(value, _min, _max):
    callback = model.parse_int(_min, _max)
    response = callback(None, value)
    assert response == int(value)


@mark.unit
@mark.parametrize(
    'value,_min,_max,_error_type,_error',
    [
        (0, 1, 2, ValueError, 'Provided value is less than the minimum allowed'),
        (4, 0, 2, ValueError, 'Provided value is greater than the maximum allowed'),
        ({}, 4, 2, TypeError, 'Value must be parsable to integer'),
    ],
)
def test_parse_int_errors(value, _min, _max, _error_type, _error):
    callback = model.parse_int(_min, _max)
    with raises(_error_type) as exc:
        callback(None, value)
    assert exc.value.args[0] == _error
    assert exc.value.args[1]['Input Value'] == str(value)


@mark.unit
def test_module_config_default():
    cfg = model.ModuleConfig()
    assert cfg.max_generator_retries == 20
    assert cfg.max_historian_buffer == 20
    assert cfg.generator_dir == model.ADD_GENERATOR_DIR


@mark.unit
def test_module_config_max_generator_retries():
    cfg = model.ModuleConfig(**{'max_generator_retries': 10})
    assert cfg.max_generator_retries == 10
    assert cfg.max_historian_buffer == 20
    assert cfg.generator_dir == model.ADD_GENERATOR_DIR


@mark.unit
def test_module_config_max_historian_buffer():
    cfg = model.ModuleConfig(**{'max_historian_buffer': 10})
    assert cfg.max_generator_retries == 20
    assert cfg.max_historian_buffer == 10
    assert cfg.generator_dir == model.ADD_GENERATOR_DIR


@mark.unit
def test_module_config_generator_dir():
    cfg = model.ModuleConfig(**{'max_historian_buffer': 10})
    assert cfg.max_generator_retries == 20
    assert cfg.max_historian_buffer == 10
    assert cfg.generator_dir == model.ADD_GENERATOR_DIR


@mark.unit
def test_module_config_bad_dir():
    with raises(ValueError):
        model.ModuleConfig(generator_dir='not_real')


@mark.unit
def test_project_config_invalid_type():
    with raises(TypeError) as exc:
        model.ModuleConfig(max_generator_retries=[])
    assert exc.value.args[0] == 'Value must be parsable to integer', (
        'TypeError message should match output from module'
    )



from syntrend.config import model
from pytest import fixture


@fixture(scope='function')
def project(monkeypatch):
    def _config(module, config):
        project_config = model.ProjectConfig(**{'objects': {'test': config}})
        monkeypatch.setattr(module, 'CONFIG', project_config)

    return _config



from syntrend.generators import register, PropertyGenerator


@register
class IntegerGenerator(PropertyGenerator):
    type = int
    name = 'integer'
    default_config = {
        'min_offset': -500,
        'max_offset': 500,
    }
    required_modules = ['random']

    def load_kwargs(self, kwargs: dict[str, any]) -> dict[str, any]:
        kwargs['min_offset'] = int(kwargs['min_offset'])
        kwargs['max_offset'] = int(kwargs['max_offset'])
        return kwargs

    def validate(self):
        if self.kwargs.min_offset > self.kwargs.max_offset:
            raise ValueError('Min Offset must be less than or equal to Max Offset')

    def generate(self):
        return self.modules.random.randint(
            self.kwargs.min_offset, self.kwargs.max_offset
        )



from syntrend.formatters import ascii_table, Event, Collection

from pytest import mark


@mark.unit
def test_single_string(project, monkeypatch):
    project(ascii_table, {'type': 'string'})
    formatter = ascii_table.table_formatter('test')
    output = formatter(Collection(Event('generated_string')))
    assert output[0] == ' generated_string ', 'Should generate a quoted string'


@mark.unit
def test_single_number(project, monkeypatch):
    project(ascii_table, {'type': 'integer'})
    formatter = ascii_table.table_formatter('test')
    output = formatter(Collection(Event(10)))
    assert output[0] == '    10 ', 'Should generate an encoded number string'


@mark.unit
def test_single_object(project, monkeypatch):
    project(ascii_table, {'type': 'object'})
    formatter = ascii_table.table_formatter('test')
    output = formatter(Collection(Event({'f1': 'string', 'f2': 10})))
    assert output[0] == ' string 10 ', (
        'Should generate an encoded ascii_table string of multiple values'
    )


@mark.unit
def test_single_object_w_col_sep(project, monkeypatch):
    project(ascii_table, {'type': 'object', 'output': {'column_separator': '|'}})
    formatter = ascii_table.table_formatter('test')
    output = formatter(Collection(Event({'f1': 'string', 'f2': 10})))
    assert output[0] == ' string | 10 ', (
        'Should generate an encoded ascii_table string of multiple values'
    )


@mark.unit
def test_multiple_objects(project, monkeypatch):
    project(ascii_table, {'type': 'object'})
    formatter = ascii_table.table_formatter('test')
    output = formatter(
        Collection(
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
        )
    )
    assert output[0] == ' string 10 ', (
        'Should generate an encoded ascii_table string of multiple values'
    )
    assert len(output) == 3, (
        'Should generate 3 lines of ascii_table output with an extra empty line'
    )


@mark.unit
def test_multiple_objects_as_collection(project, monkeypatch):
    project(ascii_table, {'type': 'object', 'output': {'collection': True}})
    formatter = ascii_table.table_formatter('test')
    output = formatter(
        Collection(
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
        )
    )
    assert output[0] == ' f1     f2 ', 'First line should include a list encapsulation'
    assert output[1] == '===========', 'Second line should be the header separator'
    assert output[2] == ' string 10 ', (
        'Should generate an indented ascii_table string of multiple values'
    )
    assert len(output) == 5, (
        'Should generate 3 lines of an asciitable, header and header separator'
    )


@mark.unit
def test_multiple_objects_as_collection_w_col_sep(project, monkeypatch):
    project(
        ascii_table,
        {'type': 'object', 'output': {'collection': True, 'column_separator': '|'}},
    )
    formatter = ascii_table.table_formatter('test')
    output = formatter(
        Collection(
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
        )
    )
    assert output[0] == ' f1     | f2 ', (
        'First line should include a list encapsulation'
    )
    assert output[1] == '=============', 'Second line should be the header separator'
    assert output[2] == ' string | 10 ', (
        'Should generate an indented ascii_table string of multiple values'
    )
    assert len(output) == 5, (
        'Should generate 3 lines of an asciitable, header and header separator'
    )


@mark.unit
def test_multiple_objects_as_collection_w_row_sep(project, monkeypatch):
    project(
        ascii_table,
        {'type': 'object', 'output': {'collection': True, 'row_separator': '-'}},
    )
    formatter = ascii_table.table_formatter('test')
    output = formatter(
        Collection(
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
        )
    )
    assert output[0] == ' f1     f2 ', 'First line should include a list encapsulation'
    assert output[1] == '===========', 'Second line should be the header separator'
    assert output[2] == ' string 10 ', (
        'Should generate an ascii_table string of multiple values'
    )
    assert output[3] == '-----------', 'Lines between rows should use the row separator'
    assert len(output) == 8, (
        'Should generate 3 lines of an asciitable, 2 lines with row separator, header and header separator'
    )



from syntrend.generators import register, PropertyGenerator

from random import randint


@register
class StringGenerator(PropertyGenerator):
    """String Generator

    Keyword Args:
        min_length (:obj:`int`): Minimum length for a random string. Default is `6`
        max_length (:obj:`int`): Maximum length for a random string. Default is `20`
        chars (:obj:`list[str] | str`)::
            List of characters to be used for random string generation. **Changing order or adding
            duplicates will impact results**. Default is the sequence: "[0-9a-zA-Z]".

    Raises:
        ValueError::
            - Length of :attr:`chars` is not greater than one
            - :attr:`min_length` is less than 1
            - :attr:`max_length` is less than :attr:`min_length`

    Notes:
        - Values in :attr:`chars` must be parseable as a string object
        - :attr:`chars` can be a list
            - items can be individual characters or multi-character words. Something like the following can create a sequence like `"is bigthisis biglist"`

                type: string
                min_length: 3
                max_length: 6
                chars:
                    - this
                    - list
                    - is big

    """

    name = 'string'
    type = str
    default_config = {
        'chars': '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ',
        'min_length': 6,
        'max_length': 20,
    }

    def load_kwargs(self, kwargs: dict[str, any]) -> dict[str, any]:
        kwargs['min_length'] = int(kwargs['min_length'])
        kwargs['max_length'] = int(kwargs['max_length'])
        return kwargs

    def validate(self):
        if len(self.kwargs.chars) <= 0:
            raise ValueError(
                'Cannot generate random strings without a list of characters'
            )
        if self.kwargs.min_length <= 0:
            raise ValueError(
                'Min Length for string generator must be greater than zero'
            )
        if self.kwargs.min_length > self.kwargs.max_length:
            raise ValueError('Min Length must be less than or equal to Max length')

    def generate(self):
        return ''.join(
            [
                self.kwargs.chars[randint(0, len(self.kwargs.chars) - 1)]
                for _ in range(randint(self.kwargs.min_length, self.kwargs.max_length))
            ]
        )


@register
class HexGenerator(StringGenerator):
    name = 'hex'
    default_config = {
        'use_upper': False,
        'max_length': 20,
        'min_length': 6,
        'chars': '0123456789abcdef',
    }

    def load_kwargs(self, kwargs: dict[str, any]) -> dict[str, any]:
        kwargs['min_length'] = int(kwargs['min_length'])
        kwargs['max_length'] = int(kwargs['max_length'])
        kwargs['char_length'] = kwargs['max_length'] - kwargs['min_length'] - 1
        kwargs['use_upper'] = bool(kwargs['use_upper'])
        if kwargs['use_upper']:
            kwargs['chars'] = kwargs['chars'].upper()
        return kwargs

    def validate(self):
        assert self.kwargs.min_length <= self.kwargs.max_length, (
            'Min Length must be less than or equal to Max Length'
        )
        assert self.kwargs.char_length >= 0, 'Cannot generate string from an empty list'

    def generate(self):
        return ''.join(
            [
                self.kwargs.chars[randint(0, self.kwargs.char_length)]
                for _ in range(randint(self.kwargs.min_length, self.kwargs.max_length))
            ]
        )



from syntrend.generators import register, PropertyGenerator

from faker import Faker

fake = Faker()


@register
class NameGenerator(PropertyGenerator):
    type = str
    name = 'name'

    def generate(self):
        return fake.name()


@register
class FirstNameGenerator(PropertyGenerator):
    type = str
    name = 'first_name'

    def generate(self):
        return fake.first_name()


@register
class LastNameGenerator(PropertyGenerator):
    type = str
    name = 'last_name'

    def generate(self):
        return fake.last_name()



from syntrend.formatters import register_formatter, Collection
from syntrend.config import CONFIG

DEFAULT_LINESEP = '\r\n'


@register_formatter('csv')
def csv_formatter(object_name: str):
    import csv
    from io import StringIO

    buffer = StringIO()
    csv_writer = csv.DictWriter(buffer, {}, quoting=csv.QUOTE_NONNUMERIC)
    output_options = CONFIG.objects[object_name].output

    def _write_events(events: list[dict[str, object]]):
        for event in events:
            csv_writer.writerow(event)

    def __formatter(event: Collection) -> list[str]:
        buffer.seek(0)
        buffer.truncate()

        if not csv_writer.fieldnames:
            csv_writer.fieldnames = list(event[0])
        if output_options.collection:
            csv_writer.writeheader()
        _write_events(event)
        return buffer.getvalue().split(DEFAULT_LINESEP)

    return __formatter



from syntrend.config import CONFIG
from syntrend.formatters import register_formatter, Collection

SQL_INSERT_FORMAT = 'insert into {table} ({columns}) values ({values})'


@register_formatter('sql')
def sql_formatter(object_name: str):
    _ = CONFIG.objects[object_name].output  # To satisfy testing and lint requirements

    def __format_value(value):
        if isinstance(value, str):
            return f'"{str(value)}"'
        return str(value)

    def __format_event(event: dict) -> str:
        return SQL_INSERT_FORMAT.format(
            table=object_name,
            columns=', '.join(list(event)),
            values=', '.join(map(__format_value, event.values())),
        )

    def __formatter(events: Collection) -> list[str]:
        return list(
            map(
                __format_event,
                [ev if isinstance(ev, dict) else {'value': ev} for ev in events],
            )
        )

    return __formatter



from syntrend.config import CONFIG
from syntrend.utils import writers

from pathlib import Path
from tempfile import mkstemp
from typing import Callable, Union
from os import close, linesep
from importlib import import_module
import pickle
import logging
import sys
from collections import namedtuple

LOG = logging.getLogger(__name__)
T_Primary = Union[str, int, float, dict[str, 'T_Primary']]
Formatter = namedtuple('Formatter', 'format,close')
T_Formatter = Callable[[str], Callable[['Collection'], list[str]]]
FORMATTERS: dict[str, T_Formatter] = {}

STREAM_FEED_EVENT = 'stream_event'
STREAM_FEED_COLLECTION = 'stream_collection'
TH_TEMP_PATH = 'temp_path'

FORMAT_FIELDS_DEFAULT = {
    'nl': linesep,
}


class Event(dict):
    def __init__(self, value):
        self.use_default = False
        if not isinstance(value, dict):
            value = {'value': value}
            self.use_default = True
        super().__init__(**value)


class Collection(list):
    def __init__(self, *events: Event):
        super().__init__(events)


class TempHandler(writers.OutputHandler):
    def load(self):
        fd, temp_path = mkstemp(prefix='syntrend_')
        close(fd)
        self.args[TH_TEMP_PATH] = Path(temp_path)
        self.args['output_callback'] = lambda x: None

    def write(self, event: Event, _=False):
        if (
            self.args[TH_TEMP_PATH].exists()
            and self.args[TH_TEMP_PATH].stat().st_size > 0
        ):
            with open(self.args[TH_TEMP_PATH], 'rb') as t_file:
                _buffer = pickle.load(t_file)
        else:
            _buffer = Collection()
        _buffer.append(event)
        with open(self.args[TH_TEMP_PATH], 'wb') as t_file:
            pickle.dump(_buffer, t_file)

    def get_collection(self) -> Collection:
        with open(self.args[TH_TEMP_PATH], 'r+b') as t_file:
            return pickle.load(t_file)

    def close(self):
        self.args[TH_TEMP_PATH].unlink(missing_ok=True)


def register_formatter(format_name: str):
    if format_name in FORMATTERS:
        raise NameError(
            'Formatter is already registered', {'Formatter Name': format_name}
        )

    def _register_formatter(func: T_Formatter):
        FORMATTERS[format_name] = func
        return func

    return _register_formatter


def load_formatter(object_name: str) -> Formatter:
    output_config = CONFIG.objects[object_name].output
    output_handler = writers.setup_event_stream(object_name)
    formatter = FORMATTERS[output_config.format](object_name)
    if output_config.collection:
        temp_collector = TempHandler(object_name)
        temp_collector.load()
    else:
        temp_collector = writers.OutputHandler(object_name)

    def __handle_events(event: dict):
        clear_target = False
        event = Event(event)
        if output_config.collection:
            temp_collector.write(event)
            if isinstance(output_handler, writers.ConsoleHandler):
                return
            clear_target = True
            event = temp_collector.get_collection()
        else:
            event = Collection(event)
        formatted_output = formatter(event)
        output_handler.write(linesep.join(formatted_output) + linesep, clear_target)

    def __handle_close():
        if output_config.collection:
            events = temp_collector.get_collection()
            formatted_output = formatter(events)
            output_handler.write(linesep.join(formatted_output) + linesep)
        temp_collector.close()
        output_handler.close()

    return Formatter(__handle_events, __handle_close)


def _load_formatter_dir(module_name: str, directory: Path):
    for _file in directory.iterdir():
        if (
            not _file.suffix.startswith('.py')
            or _file.is_dir()
            or _file.name.startswith('_')
        ):
            continue
        basename = _file.name.split('.')[0]
        _ = import_module(f'{module_name}.{basename}')


def load_formatters():
    _load_formatter_dir('syntrend.formatters', Path(__file__).parent)
    add_formatter_pkg = Path(CONFIG.config.generator_dir).absolute()
    if not (add_formatter_pkg.is_dir() and add_formatter_pkg.exists()):
        return
    formatter_pkg_name = add_formatter_pkg.name
    sys.path.append(str(add_formatter_pkg))
    _load_formatter_dir(formatter_pkg_name, add_formatter_pkg)



from syntrend.formatters import csv, Event, Collection

from pytest import mark


@mark.unit
def test_single_string(project, monkeypatch):
    project(csv, {'type': 'string'})
    formatter = csv.csv_formatter('test')
    output = formatter(Collection(Event('generated_string')))
    assert output[0] == '"generated_string"', 'Should generate a quoted string'


@mark.unit
def test_single_number(project, monkeypatch):
    project(csv, {'type': 'integer'})
    formatter = csv.csv_formatter('test')
    output = formatter(Collection(Event(10)))
    assert output[0] == '10', 'Should generate an encoded number string'


@mark.unit
def test_single_object(project, monkeypatch):
    project(csv, {'type': 'object'})
    formatter = csv.csv_formatter('test')
    output = formatter(Collection(Event({'f1': 'string', 'f2': 10})))
    assert output[0] == '"string",10', (
        'Should generate an encoded csv string of multiple values'
    )


@mark.unit
def test_multiple_objects(project, monkeypatch):
    project(csv, {'type': 'object'})
    formatter = csv.csv_formatter('test')
    output = formatter(
        Collection(
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
        )
    )
    assert output[0] == '"string",10', (
        'Should generate an encoded csv string of multiple values'
    )
    assert len(output) == 4, (
        'Should generate 3 lines of csv output with an extra empty line'
    )


@mark.unit
def test_multiple_objects_as_collection(project, monkeypatch):
    project(csv, {'type': 'object', 'output': {'collection': True}})
    formatter = csv.csv_formatter('test')
    output = formatter(
        Collection(
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
            Event({'f1': 'string', 'f2': 10}),
        )
    )
    assert output[0] == '"f1","f2"', 'First line should include a header'
    assert output[1] == '"string",10', (
        'Should generate an encoded csv string of multiple values'
    )
    assert len(output) == 5, (
        'Should generate 4 lines of csv output with header, and an extra empty line'
    )



from syntrend.config import model, CONFIG
from syntrend.utils import distributions, exc
from typing import Type, Callable
from pathlib import Path
from importlib import import_module
from collections import namedtuple
import logging

from syntrend.generators.__base_types import load_type

LOG = logging.getLogger(__name__)
GENERATORS: dict[str, Type['PropertyGenerator']] = {}


def default_generator(new, **kwargs):
    return new


class PropertyGenerator:
    """Base Property Generator for all generators

    Keyword Args:
        type (:obj:`str`): Defines the generator to use. See sections below for options
        start: A Starting Value of any type. Necessary if building an expression using a previous value.
        expression (:obj:`str`): Parsable expression (using `Jinja <https://jinja.palletsprojects.com/en/3.1.x/>`__) used to generate expected values. More information in `Expressions Doc </expressions>`__)
    """

    type: Type = None
    name: str = ''
    default_config: dict[str, any] = {}
    required_modules: list[str] = []

    def __init__(self, object_name: str, config: model.PropertyDefinition):
        self.root_object = object_name
        self.config = config
        self.properties: dict[str, any] = {}
        self.items: list[any] = []
        self.expression: Callable = default_generator
        self.start = None
        self.__distribution = None

        kwargs_names = list(self.config.kwargs)
        __modules_nt_type = namedtuple('RequiredModules', self.required_modules)
        __kwargs_nt_type = namedtuple(self.name, kwargs_names)
        self.modules: __modules_nt_type = __modules_nt_type(*self.required_modules)
        self.kwargs: __kwargs_nt_type = __kwargs_nt_type(**self.config.kwargs)
        self.root_manager = None
        self.iteration = -1
        self.iteration_value = None
        self.__modules_nt_type = __modules_nt_type
        self.__kwargs_nt_type = __kwargs_nt_type

    def __repr__(self):
        return f'<{self.__class__.__name__}: current={self.iteration_value}>'

    def load(self, manager):
        self.root_manager = manager
        self.modules = self.__modules_nt_type(
            **{mod_name: import_module(mod_name) for mod_name in self.required_modules}
        )
        kwargs = self.load_kwargs(self.config.kwargs)
        kwargs_tpl = namedtuple(self.name, list(kwargs))
        self.start = self.config.start
        self.kwargs = kwargs_tpl(**kwargs)
        self.properties = self.load_properties(self.config.properties)
        self.items = self.load_items(self.config.items)
        self.validate()
        self.__distribution = distributions.get_distribution(self.config.distribution)
        if self.config.expression and isinstance(self.config.expression, str):
            self.expression = manager.load_expression(self)

    def load_items(self, items: list[any]) -> list[any]:
        return items

    def load_properties(self, properties: dict[str, any]) -> dict[str, any]:
        return properties

    def load_kwargs(self, kwargs: dict[str, any]) -> dict[str, any]:
        return kwargs

    def validate(self):
        pass

    def render(self):
        iteration = self.root_manager.current_iteration(self.root_object)
        if self.iteration == iteration:
            return self.iteration_value

        self.iteration = iteration
        if not iteration and self.start is not None:
            self.iteration_value = self.start
            return self.start

        generated = self.generate()
        try:
            self.iteration_value = self.expression(
                new=generated, interval=self.iteration, kwargs=self.kwargs
            )
        except (ValueError, TypeError) as e:
            exc.process_exception(e)
        self.iteration_value = self.__distribution(self.iteration_value)
        if self.type is not None and not isinstance(self.iteration_value, self.type):
            self.iteration_value = self.type(self.iteration_value)
        return self.iteration_value

    def undo(self):
        self.iteration -= 1

    def generate(self):
        raise NotImplementedError('Generator has not implemented `generate` method')


def register(property_generator: Type[PropertyGenerator]):
    assert property_generator.name, 'Property Generator must have a name specified'
    assert property_generator.name not in GENERATORS, (
        f"Property Generator '{property_generator.name}' already registered"
    )
    GENERATORS[property_generator.name] = load_type(property_generator)
    return property_generator


def get_generator(
    object_name: str, config: model.PropertyDefinition, manager
) -> PropertyGenerator:
    prop_gen_cls = GENERATORS[config.type]
    new_config = model.PropertyDefinition(
        name=config.name, type=config.type, **prop_gen_cls.default_config
    )
    model.update(new_config, config)
    new_gen = prop_gen_cls(object_name, new_config)
    new_gen.load(manager)
    return new_gen


def _load_generator_dir(module_name: str, directory: Path):
    for _file in directory.iterdir():
        if (
            not _file.suffix.startswith('.py')
            or _file.is_dir()
            or _file.name.startswith('_')
        ):
            continue
        basename = _file.name.split('.')[0]
        _ = import_module(f'{module_name}.{basename}')


def load_generators():
    _load_generator_dir('syntrend.generators', Path(__file__).parent)
    add_generator_pkg = Path(CONFIG.config.generator_dir).absolute()
    if not (add_generator_pkg.is_dir() and add_generator_pkg.exists()):
        return
    import sys

    generator_pkg_name = add_generator_pkg.name
    sys.path.append(str(add_generator_pkg))
    _load_generator_dir(generator_pkg_name, add_generator_pkg)



from syntrend.generators import register, PropertyGenerator


@register
class UUIDGenerator(PropertyGenerator):
    type = str
    name = 'uuid'
    default_config = {
        'use_upper': False,
        'compact': False,
        'separator': '-',
    }
    required_modules = ['uuid']

    def load_kwargs(self, kwargs: dict[str, any]) -> dict[str, any]:
        kwargs['use_upper'] = bool(kwargs['use_upper'])
        kwargs['compact'] = bool(kwargs['compact'])
        kwargs['separator'] = str(kwargs['separator'])
        return kwargs

    def generate(self):
        uuid_val = self.modules.uuid.uuid4()
        if self.kwargs.compact:
            uuid_val = uuid_val.hex
        uuid_val = str(uuid_val).replace('-', self.kwargs.separator)
        if self.kwargs.use_upper:
            uuid_val = uuid_val.upper()
        return uuid_val



from syntrend.generators import register, PropertyGenerator

from datetime import datetime, timedelta, timezone
from time import localtime
from functools import partial
import re

RE_TIME_OFFSET = re.compile(r'(-?\d+)([YmdHMS])')
DATETIME_SYMBOL_MAP = {
    'Y': 'years',
    'm': 'months',
    'd': 'days',
    'H': 'hours',
    'M': 'minutes',
    'S': 'seconds',
}


def datetime_aware(kwargs):
    kwargs['is_utc'] = bool(kwargs.get('is_utc', True))
    kwargs['tz_offset'] = timezone(
        timedelta(seconds=0 if kwargs['is_utc'] else localtime().tm_gmtoff)
    )
    kwargs['min_offset'] = int(kwargs.get('min_offset', 0))
    kwargs['max_offset'] = int(kwargs.get('max_offset', kwargs['min_offset']))
    if kwargs['min_offset'] > kwargs['max_offset']:
        raise ValueError('min_offset must be less than or equal to max_offset')

    def generator(self):
        return self.kwargs.now().astimezone(self.kwargs.tz_offset)

    kwargs['generate'] = generator
    kwargs['now'] = partial(datetime.now, tz=timezone.utc)


@register
class DateTimeGenerator(PropertyGenerator):
    type = str
    name = 'datetime'
    default_config = {
        'format': '%Y-%m-%dT%H:%M:%S%z',
        'is_utc': True,
        'time_offset': '0d',
    }

    def load_kwargs(self, kwargs: dict[str, any]) -> dict[str, any]:
        offset_matches = {}
        for _match in RE_TIME_OFFSET.finditer(kwargs['time_offset']):
            val, key = _match.groups()
            offset_matches[DATETIME_SYMBOL_MAP[key]] = int(val)
        kwargs['time_offset'] = timedelta(**offset_matches)
        datetime_aware(kwargs)
        return kwargs

    def generate(self):
        return (self.kwargs.generate(self) + self.kwargs.time_offset).strftime(
            self.kwargs.format
        )


@register
class TimestampGenerator(PropertyGenerator):
    type = int
    name = 'timestamp'
    default_config = {
        'time_offset': 0,
    }

    def load_kwargs(self, kwargs: dict[str, any]) -> dict[str, any]:
        datetime_aware(kwargs)
        kwargs['time_offset'] = int(kwargs['time_offset'])
        return kwargs

    def generate(self):
        return int(self.kwargs.generate(self).timestamp()) + self.kwargs.time_offset



import datetime
import random
import re
import logging
from typing import Union, TYPE_CHECKING
import math

if TYPE_CHECKING:
    from syntrend.utils.manager import SeriesManager

LOG = logging.getLogger(__name__)
R_DATETIME = re.compile(r'[1-2]\d{3}-[0-1]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-5]\d')
R_TIME = re.compile(r'[0-2]\d:[0-5]\d:[0-5]\d')
R_DATE = re.compile(r'[1-2]\d{3}-[0-1]\d-[0-3]\d')
R_DELTA = re.compile(r'(\d+)([dHMS])')
DELTA_ABBR_MAP = {'d': 'days', 'H': 'hours', 'M': 'minutes', 'S': 'seconds'}

MANAGER: Union['SeriesManager', None] = None


def to_timestamp(value: datetime.datetime):
    return value.timestamp()


def to_datetime(value_string, format_str: str = None):
    for formatter in [
        datetime.datetime.fromisoformat,
        datetime.date.fromisoformat,
        datetime.time.fromisoformat,
    ]:
        try:
            return formatter(value_string)
        except (TypeError, ValueError) as e:
            print(e)
            pass
    if format_str is not None:
        try:
            return datetime.datetime.strptime(value_string, format_str)
        except ValueError:
            pass
    time_parts = {}
    for match in R_DELTA.finditer(value_string):
        time_parts[DELTA_ABBR_MAP[match.group(2)]] = int(match.group(1))
    return datetime.timedelta(**time_parts)


def series(generator, series_length):
    class _Comparator:
        def __init__(self):
            self._v = [
                generator.condition_historian(i) for i in range(0, -series_length, -1)
            ]

        def __eq__(self, other):
            return all([a == other for a in self._v])

        def __ne__(self, other):
            return all([a != other for a in self._v])

        def __gt__(self, other):
            return all([a > other for a in self._v])

        def __ge__(self, other):
            return all([a >= other for a in self._v])

        def __lt__(self, other):
            return all([a < other for a in self._v])

        def __le__(self, other):
            return all([a <= other for a in self._v])

        def __contains__(self, item):
            return item in self._v

    return _Comparator()


def get_object(object_name):
    historian = MANAGER.historians[object_name]
    generator = MANAGER.generators[object_name]

    def get_index(index: int = 0):
        if index == 0:
            return generator
        if index <= len(historian):
            return historian[index - 1]
        return None

    return get_index


def load_environment(manager: 'SeriesManager'):
    global MANAGER
    MANAGER = manager
    manager.expression_env.globals.update(
        datetime=datetime.datetime,
        timedelta=datetime.timedelta,
        time=datetime.time,
        date=datetime.date,
        sin=math.sin,
        cos=math.cos,
        tan=math.tan,
        degrees=math.degrees,
        radians=math.radians,
        random=random.randint,
    )
    manager.expression_env.filters.update(
        to_timestamp=to_timestamp,
        to_datetime=to_datetime,
        series=series,
        # path=get_path,
    )
    for object_name in manager.historians:
        manager.expression_env.globals[object_name] = get_object(object_name)



from syntrend.config import model

from random import random, gauss


def dist_no_dist(_):
    def _generate(input_value):
        return input_value

    return _generate


def dist_linear(prop_dist: model.PropertyDistribution):
    scale = prop_dist.max_offset - prop_dist.min_offset

    def _generator(input_value):
        return input_value + random() * scale + prop_dist.min_offset

    return _generator


def dist_standard_deviation(prop_dist: model.PropertyDistribution):
    # if prop_dist.min_offset >= prop_dist.max_offset:
    #     raise ValueError("min_offset must be less than max_offset")
    # scale = prop_dist.max_offset - prop_dist.min_offset
    # unscaled_variance = (float(prop_dist.std_dev_factor) / scale) ** 2
    # unscaled_mean = -prop_dist.min_offset / scale
    # t = unscaled_mean / (1 - unscaled_mean)
    # beta = ((t / unscaled_variance) - (t * t) - (2 * t) - 1) / (
    #     (t * t * t) + (3 * t * t) + (3 * t) + 1
    # )
    # alpha = beta * t
    # if alpha <= 0 or beta <= 0:
    #     raise ValueError("Cannot create value in Std Dev with given numbers", alpha, beta)

    def _generator(input_value):
        return gauss(input_value, float(prop_dist.std_dev))

    return _generator


DISTRIBUTIONS = {
    model.DistributionTypes.NoDistribution: dist_no_dist,
    model.DistributionTypes.Linear: dist_linear,
    model.DistributionTypes.StandardDeviation: dist_standard_deviation,
}


def get_distribution(prop_def: model.PropertyDistribution):
    return DISTRIBUTIONS[prop_def.type](prop_def)



from syntrend.generators import register, PropertyGenerator, get_generator
from syntrend.config.model import PropertyDefinition

from random import randint


class BaseComplexGenerator(PropertyGenerator):
    def get_children(self) -> list[str]:
        raise NotImplementedError(
            f"{str(type(self))}: 'get_children' is not implemented"
        )


@register
class UnionGeneratorBase(BaseComplexGenerator):
    """Generates a single value based on a random selection of many provided generator options

    Args:
        items (list[`:obj:PropertyGenerator`]): List of generator types with their own properties

    Example:
        Selectively choose a generator randomly or through the use of an expression::

            type: union
            items:
                - type: integer
                  min_offset: -4
                  max_offset: 4
                - type: integer
                  min_offset: 10
                  max_offset: 20
    """

    name = 'union'

    def get_children(self):
        return []

    def load_items(self, items: list[any]) -> list[any]:
        _gens = []
        for item in items:
            _gens.append(
                get_generator(
                    self.root_object, PropertyDefinition(**item), self.root_manager
                )
            )
        return _gens

    def generate(self) -> any:
        return self.items[randint(0, len(self.items) - 1)].render()


@register
class ListGeneratorBase(BaseComplexGenerator):
    """Generates a list-object of 1 or many values of the specified generator type"""

    name = 'list'
    default_config = {
        'min_length': 1,
        'max_length': 5,
    }

    def get_children(self):
        return []

    def load_kwargs(self, kwargs):
        kwargs['min_length'] = int(kwargs['min_length'])
        kwargs['max_length'] = int(kwargs['max_length'])
        assert kwargs['min_length'] <= kwargs['max_length'], (
            'Min must be less than or equal to Max'
        )
        assert kwargs['min_length'] >= 0, 'List cannot have a negative length'
        assert 'sub_type' in kwargs, (
            "Must provide a 'sub_type' property for the values to be generated"
        )
        kwargs['sub_type'] = get_generator(
            self.root_object,
            PropertyDefinition(**kwargs['sub_type']),
            self.root_manager,
        )
        return kwargs

    def generate(self) -> list[any]:
        return [
            self.kwargs.sub_type.generate()
            for _ in range(randint(self.kwargs.min_length, self.kwargs.max_length))
        ]


@register
class ObjectGeneratorBase(BaseComplexGenerator):
    """Generates an object based on a defined mapping of properties"""

    name = 'object'
    type = dict

    def get_children(self) -> list[str]:
        return list(self.properties)

    def load_properties(
        self, properties: dict[str, any]
    ) -> dict[str, PropertyGenerator]:
        return {
            key: get_generator(self.root_object, properties[key], self.root_manager)
            for key in properties
        }

    def generate(self):
        for key in self.properties:
            self.properties[key].render()
        return {key: self.properties[key].render() for key in self.properties}

    def undo(self):
        super(BaseComplexGenerator, self).undo()
        for key in self.properties:
            self.properties[key].undo()



from syntrend.config.model import ProjectConfig, copy, update
from syntrend.config.parse import load_config as _load_config

default_config = {'objects': {'this': {'type': 'integer'}}}
CONFIG = ProjectConfig(**default_config)


def load_config(config_file):
    global CONFIG
    _cfg = _load_config(config_file)
    CONFIG.objects = _cfg.objects
    CONFIG.config = _cfg.config
    CONFIG.output = _cfg.output

    for obj_name in CONFIG.objects:
        _output = copy(CONFIG.output)
        update(_output, CONFIG.objects[obj_name].output)
        CONFIG.objects[obj_name].output = _output


__all__ = [
    ProjectConfig,
    load_config,
]



from syntrend.config import CONFIG

from typing import Any, Callable
from pathlib import Path
import sys
from os import linesep

T_Event_Callback = Callable[[None], Any]
BASE_OUTPUT_CONFIG = CONFIG.output
STREAM_TARGET_FILE = 'stream_file'
STREAM_TARGET_CONSOLE = 'stream_console'
CONSOLE_DEFAULT_EVENT_FORMAT = '{body}'
CONSOLE_DEFAULT_COLLECTION_FORMAT = '{body}'


class OutputHandler:
    def __init__(self, object_name: str):
        self.object_def = CONFIG.objects[object_name]
        self.object_name = object_name
        self.args = {}

    def load(self):
        return

    def write(self, event: str, clear=False) -> None:
        return

    def close(self) -> None:
        return


class ConsoleHandler(OutputHandler):
    def load(self):
        args = {
            'name': self.object_name,
            'buffer': '',
            'body': '{body}',
            'nl': linesep,
        }
        if self.object_def.output.collection:
            out_format = self.object_def.output.kwargs.get(
                'console_collection_format',
                CONSOLE_DEFAULT_COLLECTION_FORMAT,
            )
        else:
            longest_name = max([len(obj_name) for obj_name in CONFIG.objects])
            args['buffer'] = ' ' * (longest_name - len(self.object_name) + 1)
            out_format = self.object_def.output.kwargs.get(
                'console_event_format', CONSOLE_DEFAULT_EVENT_FORMAT
            )
        self.args['format'] = out_format.format(**args)

    def write(self, content: str, clear=False) -> None:
        sys.stdout.write(self.args['format'].format(body=content))


class FileHandler(OutputHandler):
    def load(self):
        self.args['sequence'] = 0
        self.args['path'] = Path(self.object_def.output.directory).joinpath(
            self.object_def.output.filename_format.format(
                name=self.object_name,
                format=self.object_def.output.format,
                id='{id}',
            )
        )

    def write(self, content: str, clear=False):
        self.args['sequence'] += 1
        file_path = Path(str(self.args['path']).format(id=self.args['sequence']))
        if clear:
            file_path.unlink(missing_ok=True)
        if not file_path.exists():
            file_path.open(mode='x')
        with file_path.open(mode='r+b') as f:
            f.write(content.encode('utf-8'))


def setup_event_stream(object_name: str) -> OutputHandler:
    output_config = CONFIG.objects[object_name].output
    if output_config.directory and str(output_config.directory) != '-':
        handler = FileHandler(object_name)
    else:
        handler = ConsoleHandler(object_name)
    handler.load()

    return handler



from random import randint
from syntrend.generators import register, PropertyGenerator


@register
class ChoiceGenerator(PropertyGenerator):
    """Generator to randomly provide one of many Generator Types for a value

    Args:
        items (list): List of static values to select from

    Example:
        Selectively choose a generator randomly or through the use of an expression::

            type: choice
            items:
                - a random string
                - 155
                - [a, list, of, 166, 215, numbers]
    """

    name = 'choice'

    def validate(self):
        assert len(self.items) > 0, 'Cannot generate items from an empty list'

    def generate(self):
        return self.items[randint(0, len(self.items) - 1)]



from syntrend.utils import filters

from datetime import datetime, timezone

from pytest import mark


@mark.unit
def test_to_timestamp():
    exp_timestamp = 1704067200
    input_datetime = datetime(2024, 1, 1, tzinfo=timezone.utc)
    result = filters.to_timestamp(input_datetime)
    assert result == exp_timestamp, 'Generated datetime should convert to timestamp'


@mark.unit
@mark.parametrize(
    'value,result',
    [
        ('2024-01-01T00:00:00Z', datetime(2024, 1, 1, tzinfo=timezone.utc)),
        ('2024-01-01+00:00', datetime(2024, 1, 1)),
        ('2024-01-01T00:00:00+00:00', datetime(2024, 1, 1, tzinfo=timezone.utc)),
    ],
)
def test_to_datetime(value, result):
    returned = filters.to_datetime(value)
    assert returned == result, 'Generated datetime should convert to datetime'


@mark.unit
@mark.parametrize(
    'delta,new_time',
    [
        ('1d', datetime(2024, 1, 2, tzinfo=timezone.utc)),
        ('1H', datetime(2024, 1, 1, hour=1, tzinfo=timezone.utc)),
        ('1M', datetime(2024, 1, 1, minute=1, tzinfo=timezone.utc)),
        ('1S', datetime(2024, 1, 1, second=1, tzinfo=timezone.utc)),
        ('1d 5H', datetime(2024, 1, 2, hour=5, tzinfo=timezone.utc)),
        ('1d 1H 1M', datetime(2024, 1, 2, hour=1, minute=1, tzinfo=timezone.utc)),
        ('1d 5M 2d', datetime(2024, 1, 3, minute=5, tzinfo=timezone.utc)),
    ],
)
def test_to_datetime_delta(delta, new_time):
    returned = datetime(2024, 1, 1, tzinfo=timezone.utc) + filters.to_datetime(delta)
    assert returned == new_time, 'Generated datetime should convert to datetime'



import traceback
import json

from pytest import mark


@mark.docs
def test_single_value_string(project_result):
    assert project_result.exit_code == 0, 'Command should complete successfully'
    assert 10 <= len(project_result.output.splitlines()[0].strip()) <= 12, (
        'Length of string should be 8 - 10 chars long (plus quotes)'
    )


@mark.docs
def test_single_value_integer(project_result):
    assert project_result.exit_code == 0, f'Command Errored: {project_result.exception}'
    gen_val = int(project_result.output)
    assert 5 <= gen_val <= 10, 'Generated Integer should be between 5 and 10'


@mark.docs
def test_random_choice(project_result):
    assert project_result.exit_code == 0, 'Command should complete successfully'
    outputs = [line for line in project_result.output.splitlines() if line]
    assert outputs[-1] in {
        '"red"',
        '"yellow"',
        '"blue"',
        '"orange"',
        '"green"',
        '"purple"',
    }, 'Choice should be a colour'


@mark.docs
def test_single_value_object(project_result):
    if project_result.exit_code:
        traceback.print_tb(project_result.exc_info[2])
    assert project_result.exit_code == 0, (
        f'Command Errored: {project_result.exception}\n{traceback.print_stack(project_result.exc_info)}'
    )
    gen_obj = json.loads(project_result.output)
    assert isinstance(gen_obj, dict), (
        f'Expecting dict but got {type(gen_obj).__name__} instead'
    )
    assert len(gen_obj) == 4, 'Each object should have 3 properties'
    assert 'field_1' in gen_obj, "'field_1' should be an object in both sets"
    assert isinstance(gen_obj['field_1'], str)
    assert isinstance(gen_obj['field_2'], int)
    assert isinstance(gen_obj['field_3'], float)
    assert gen_obj['field_4'] in {'small', 'medium', 'large'}


@mark.docs
def test_multi_value_string(project_result):
    assert project_result.exit_code == 0, 'Command should complete successfully'
    generated_lines = project_result.output[:-1].splitlines()
    assert len(generated_lines) == 5, 'Should have generated 5 values'
    assert all([8 <= len(line) <= 22 for line in generated_lines])


@mark.docs
def test_static_ref_events(project_result):
    assert project_result.exit_code == 0, 'Command should complete successfully'
    generated_lines = [
        json.loads(line) for line in project_result.output[:-1].splitlines()
    ]
    names = [ev['user_id'] for ev in generated_lines]
    timestamps = [ev['timestamp'] for ev in generated_lines]
    assert len(generated_lines) == 5, 'Should have generated 5 values'
    assert all([name == 'jdoe' for name in names])
    assert all(
        [timestamps[a - 1] + 5 == timestamps[a] for a in range(1, len(timestamps))]
    )


@mark.docs
def test_static_ref_random_start(project_result):
    assert project_result.exit_code == 0, 'Command should complete successfully'
    generated_lines = [
        json.loads(line) for line in project_result.output[:-1].splitlines()
    ]
    names = [ev['user_id'] for ev in generated_lines]
    timestamps = [ev['timestamp'] for ev in generated_lines]
    assert len(generated_lines) == 5, 'Should have generated 5 values'
    assert 3 <= len(names[0]) <= 6
    assert all([name == names[0] for name in names])
    assert all(
        [timestamps[a - 1] + 5 == timestamps[a] for a in range(1, len(timestamps))]
    )


@mark.docs
def test_seq_format_string(project_result):
    assert project_result.exit_code == 0, 'Command should complete successfully'
    generated_lines = [
        json.loads(line) for line in project_result.output[:-1].splitlines()
    ]
    assert all([ev.endswith('-test') for ev in generated_lines])
    assert all([len(ev) == 10 for ev in generated_lines])


@mark.docs
def test_cond_status_change(project_result):
    assert project_result.exit_code == 0, 'Command should complete successfully'
    generated_lines = [
        json.loads(line) for line in project_result.output[:-1].split('\n')
    ]
    assert all([ev['ref'] == 'status' for ev in generated_lines])
    below, above = [], []
    ranges = (below, above)
    for ev in generated_lines:
        ranges[ev['sensor'] > 5].append(ev)
    assert len(below) > 1
    assert len(above) > 1
    assert all([ev['status'] == 'below' for ev in below]), (
        "Status should report 'below' when Sensor value is less than 6"
    )
    assert all([ev['status'] == 'above' for ev in above]), (
        "Status should report 'above' when sensor value is greater than 5"
    )


@mark.docs
def test_num_trend(project_result):
    assert project_result.exit_code == 0, 'Command should complete successfully'
    generated_lines = [int(line) for line in project_result.output[:-1].split('\n')]
    assert all(
        [
            generated_lines[i + 1] - 1 == generated_lines[i]
            for i in range(len(generated_lines) - 1)
        ]
    ), 'Numbers should be an increment of 1 to the previous value'


@mark.docs
def test_sine_wave(project_result):
    from math import sin

    assert project_result.exit_code == 0, 'Command should complete successfully'
    generated_lines = [float(line) for line in project_result.output[:-1].split('\n')]
    new_sine_values = [
        5 if i == 0 else (10 * sin(i * 2 / 3) + 11) for i in range(0, 10)
    ]
    assert all(
        [a == b for a, b in zip(generated_lines, new_sine_values, strict=True)]
    ), 'Value should be along the Sine Wave'



from syntrend.config import model
from syntrend.utils import distributions as d, manager as m

from functools import partial

from pytest import mark, fixture

Prop_Def = partial(model.PropertyDefinition, name='test')


@fixture(scope='function')
def manager():
    mgr = m.SeriesManager()
    mgr.load()
    return mgr


@mark.unit
@mark.parametrize(
    'prop_cfg',
    [
        {'type': 'string'},
        {'type': 'string', 'distribution': model.DistributionTypes.NoDistribution},
    ],
    ids=['default_dist', 'given_dist'],
)
def test_no_distribution(prop_cfg):
    prop_cfg = Prop_Def(**prop_cfg)
    dist_func = d.get_distribution(prop_cfg.distribution)
    assert 'x' == dist_func('x'), 'The NoDist Generator returns the value it provides'


@mark.unit
@mark.parametrize('_type_str', [['integer'], ['float']])
def test_linear_distribution_numeric(_type_str: str):
    prop_cfg = Prop_Def(
        type=_type_str,
        distribution=model.PropertyDistribution(
            type=model.DistributionTypes.Linear, min_offset=0, max_offset=5
        ),
        min_offset=0,
        max_offset=5,
    )
    dist_func = d.get_distribution(prop_cfg.distribution)
    gen_values = [dist_func(5) for _ in range(30)]
    assert all([0 <= x <= 10 for x in gen_values]), (
        'All Values must be within the current tolerance'
    )


@mark.unit
@mark.parametrize('std_dev', [0.2, 0.5, 1.0, 2.0, 5.0])
def test_std_dev_distribution(std_dev):
    prop_cfg = Prop_Def(
        type='float',
        distribution=model.PropertyDistribution(
            type=model.DistributionTypes.StandardDeviation, std_dev=std_dev
        ),
    )
    offset = std_dev * 6
    dist_func = d.get_distribution(prop_cfg.distribution)
    gen_values = [dist_func(5) for _ in range(1000)]
    assert min(gen_values) > 5 - offset and max(gen_values) < 5 + offset, (
        'All Values must be within the current tolerance'
    )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import tomli
from sphinx.application import Sphinx

pyproject_file = tomli.load(open('../pyproject.toml', 'rb'))
project_config = pyproject_file['project']
tool_config = pyproject_file['tool']

project = project_config['name']
project_copyright = '2024, Will Siddall'
author = ', '.join([author['name'] for author in project_config['authors']])
release = project_config['version']

repo_url = project_config['urls']['Repository']
doc_url = project_config['urls']['Homepage']

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.doctest',
    'sphinx.ext.graphviz',
    'matplotlib.sphinxext.plot_directive',
]

templates_path = ['_templates']
exclude_patterns = []
rst_prolog = f"""
.. |git_url| replace:: {repo_url}
.. Syntrend replace:: **{project}**
"""

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']


# -- Matplotlib plot config ---
plot_include_source = False
plot_html_show_source_link = False
plot_formats = ['png']


def setup(_app: Sphinx):
    pass



from syntrend.generators import register, PropertyGenerator


@register
class StaticGenerator(PropertyGenerator):
    name = 'static'

    def validate(self):
        if not hasattr(self.kwargs, 'value'):
            raise AttributeError("Static Generator requires a 'value' property")

    def generate(self):
        return self.kwargs.value



from syntrend.config import CONFIG
from syntrend.generators import get_generator, PropertyGenerator
from syntrend.utils import historian, filters
from syntrend.formatters import load_formatter

from jinja2 import Environment, BaseLoader, exceptions

import re
import time
from datetime import datetime
from typing import Any

RE_EXPR_PATH_FUNC = re.compile(r"path\(['\"](.*?)['\"]\)")


class SeriesManager:
    def __init__(self):
        self.generators = {}
        self.formatters = {}
        self.historians: dict[str, historian.Historian] = {}
        self.expression_env = Environment(loader=BaseLoader())
        self.__renderers: dict[str, int] = {}
        self.__expr_lookups = {}

    def current_iteration(self, object_name: str) -> int:
        return self.__renderers[object_name]

    def load_expression(self, prop_generator: PropertyGenerator):
        compiled_expr = self.expression_env.compile_expression(
            prop_generator.config.expression, undefined_to_none=False
        )

        def _generate(**kwargs):
            try:
                return compiled_expr(**kwargs)
            except exceptions.UndefinedError as e:
                field = e.args[0].split(' ')[0][1:-1]
                raise ValueError(
                    'Reference field is not defined',
                    {
                        'Property': prop_generator.config.name,
                        'Missing Field': field,
                    },
                ) from None
            except exceptions.TemplateError as e:
                raise ValueError('Expression failed to execute', *e.args) from None

        return _generate

    def load(self):
        for obj_name in CONFIG.objects:
            self.historians[obj_name] = historian.Historian()
            self.generators[obj_name] = get_generator(
                obj_name, CONFIG.objects[obj_name], ROOT_MANAGER
            )
            self.formatters[obj_name] = load_formatter(obj_name)
            self.__renderers[obj_name] = 0

        filters.load_environment(self)

    def start(self):
        next_events: dict[int, list[tuple[str, Any]]] = {}

        def _run(_obj_name: str):
            _value = self.generators[_obj_name].render()
            self.formatters[_obj_name].format(_value)
            self.historians[_obj_name].append(_value)

        def _get_next_event(_obj_name: str, _current_time: int):
            if self.__renderers[_obj_name] == CONFIG.objects[_obj_name].output.count:
                return True
            _value = self.generators[_obj_name].render()
            time_value = _value[CONFIG.objects[_obj_name].output.time_field]
            if isinstance(time_value, datetime):
                time_value = time_value.timestamp()
            if _current_time and time_value <= _current_time:
                self.generators[_obj_name].undo()
                return False
            if time_value not in next_events:
                next_events[time_value] = []
            next_events[time_value].append((_obj_name, _value))
            self.__renderers[_obj_name] += 1
            return True

        collection_objects = [
            obj_name
            for obj_name in CONFIG.objects
            if CONFIG.objects[obj_name].output.collection
        ]
        event_objects = [
            obj_name
            for obj_name in CONFIG.objects
            if not CONFIG.objects[obj_name].output.collection
        ]
        time_objects = [
            obj_name
            for obj_name in event_objects
            if CONFIG.objects[obj_name].output.time_field
        ]
        non_time_objects = [
            obj_name for obj_name in event_objects if obj_name not in time_objects
        ]

        # Render collections for references
        for obj_name in collection_objects:
            for iteration in range(CONFIG.objects[obj_name].output.count):
                self.__renderers[obj_name] = iteration
                _run(obj_name)
            self.formatters[obj_name].close()

        # Render events
        for obj_name in non_time_objects:
            for iteration in range(CONFIG.objects[obj_name].output.count):
                self.__renderers[obj_name] = iteration
                _run(obj_name)
            self.formatters[obj_name].close()

        for obj_name in time_objects:
            _get_next_event(obj_name, 0)

        if len(next_events) == 0:
            return
        current_time = min(list(next_events))
        start = time.time()
        while len(next_events) > 0:
            new_time = min(list(next_events))
            if new_time != current_time:
                time.sleep(new_time - current_time - (time.time() - start))
            start = time.time()
            for obj_name, value in next_events.pop(new_time):
                self.formatters[obj_name].format(value)
                self.historians[obj_name].append(value)
                failed_count = 0
                while _get_next_event(obj_name, new_time) is False:
                    failed_count += 1
                    if failed_count == CONFIG.config.max_generator_retries:
                        raise ValueError(
                            'Failed to generate an event with a later timestamp'
                        )
            current_time = new_time

        # Close Renderers
        for obj_name in time_objects:
            self.formatters[obj_name].close()


ROOT_MANAGER = SeriesManager()



import sys
from traceback import print_tb
from os import linesep, environ


def process_exception(e: Exception):
    lines = [f'Error Encountered: ({getattr(e, "name", type(e).__name__)})']
    for arg in e.args:
        if isinstance(arg, dict):
            for k in arg:
                lines.append(f'  | {k}: {arg[k]}')
        else:
            lines.append(f'  | {arg}')
    sys.stderr.write(linesep.join(lines) + linesep)
    if environ.get('SYNTREND_DEBUG', 0):
        print_tb(e.__traceback__)
    sys.exit(1)



from livereload import Server, shell

CMD = 'uv run poe docs'

if __name__ == '__main__':
    server = Server()
    server.watch('**/*.rst', shell(CMD), delay=1)
    server.watch('docs/**/*.md', shell(CMD), delay=1)
    server.watch('docs/**/*.py', shell(CMD), delay=1)
    server.watch('docs/_static/*', shell(CMD), delay=1)
    server.watch('docs/_templates/*', shell(CMD), delay=1)
    server.serve(root='dist')



import click
import pathlib

from syntrend import generators, formatters
from syntrend.utils import manager, exc
from syntrend.config import load_config

ERRORS = []


def error(message):
    global ERRORS
    if message not in ERRORS:
        ERRORS.append(message)


@click.group(name='app')
def app():
    pass


@app.command()
@click.argument('project_file', nargs=1, type=click.Path(exists=True))
def generate(project_file: pathlib.Path):
    """
    Uses the available project file to generate datasets
    """
    try:
        load_config(project_file)
        generators.load_generators()
        formatters.load_formatters()
        manager.ROOT_MANAGER.load()
        manager.ROOT_MANAGER.start()
    except Exception as e:
        exc.process_exception(e)


@app.command()
@click.argument('project_file', nargs=1, type=click.Path(exists=True))
def validate(config_file: pathlib.Path):
    """
    Validates and summarizes a project file
    """
    pass


if __name__ == '__main__':
    app()



from typing import Type, Callable, Union


class BaseType:
    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)

    def __str__(self):
        return str(self.render())

    def render(self):
        raise NotImplementedError(
            "Generator Class has not updated it's `render` method"
        )

    def __hash__(self):
        return hash(self.render())

    def __eq__(self, other):
        return self.render() == other

    def __ne__(self, other):
        return self.render() != other


class CollectionType(BaseType):
    def __getitem__(self, item):
        return self.render()[item]

    def __len__(self):
        return len(self.render())

    def __contains__(self, item):
        return item in self.render()


class BooleanType(BaseType):
    def __and__(self, other):
        return self.render() & other

    def __or__(self, other):
        return self.render() | other

    def __xor__(self, other):
        return self.render() ^ other

    def __rxor__(self, other):
        return other ^ self.render()

    def __invert__(self):
        return ~self.render()

    def __bool__(self):
        return bool(self.render())


class NumericType(BaseType):
    def __int__(self):
        return int(self.render())

    def __float__(self):
        return float(self.render())

    def __add__(self, other):
        return self.render() + other

    def __radd__(self, other):
        return other + self.render()

    def __sub__(self, other):
        return self.render() - other

    def __rsub__(self, other):
        return other - self.render()

    def __mul__(self, other):
        return self.render() * other

    def __rmul__(self, other):
        return other * self.render()

    def __truediv__(self, other):
        return self.render() / other

    def __rtruediv__(self, other):
        return other / self.render()

    def __floordiv__(self, other):
        return self.render() // other

    def __rfloordiv__(self, other):
        return other // self.render()

    def __mod__(self, other):
        return self.render() % other

    def __rmod__(self, other):
        return other % self.render()

    def __pow__(self, other):
        return self.render() ** other

    def __rpow__(self, other):
        return other ** self.render()

    def __lt__(self, other):
        return self.render() < other

    def __le__(self, other):
        return self.render() <= other

    def __gt__(self, other):
        return self.render() > other

    def __ge__(self, other):
        return self.render() >= other

    def __abs__(self):
        return abs(self.render())


class StringType(CollectionType):
    pass


class IntegerType(NumericType):
    def __rshift__(self, other):
        return self.render() >> other

    def __lshift__(self, other):
        return self.render() << other


class FloatType(NumericType):
    def __round__(self, n=None):
        return round(self.render(), n)


class ListType(CollectionType):
    pass


class MappingType(CollectionType):
    def __getattr__(self, name):
        return self.properties[name]


EXPRESSION_TYPES = {
    str: StringType,
    int: IntegerType,
    float: FloatType,
    bool: BooleanType,
    list: ListType,
    dict: MappingType,
}
AVAILABLE_TYPES = set(EXPRESSION_TYPES.values())
T_TYPE_MAP = dict[str, set[Union[str, Type[BaseType]]]]


def __load_type_methods(type_map: T_TYPE_MAP, root_type_name: str, type_object):
    type_name = type_object.__name__
    for prop in type_object.__dict__:
        if not isinstance(getattr(type_object, prop), Callable):
            continue
        if prop not in type_map:
            type_map[prop] = set()
        type_map[prop].add(root_type_name)
    type_map['LOADED'].add(type_name)
    for base_type in type_object.__mro__[1:]:
        base_name = base_type.__name__
        if base_name == type_name:
            continue
        if not base_name.endswith('Type') or base_name == 'BaseType':
            continue
        __load_type_methods(type_map, root_type_name, base_type)


def load_type_method_mapping() -> dict[str, set[Type[BaseType]]]:
    type_methods = {'LOADED': set()}
    for _type in EXPRESSION_TYPES:
        __load_type_methods(
            type_methods, EXPRESSION_TYPES[_type].__name__, EXPRESSION_TYPES[_type]
        )
    type_methods.pop('LOADED')
    return type_methods


class AnyTime:
    def __init__(self, generator):
        self.generator = generator

    def __getattribute__(self, item):
        pass


TYPE_MAPPING = load_type_method_mapping()


def load_type(generator):
    if generator.type is None:
        return generator
    base_type = EXPRESSION_TYPES[generator.type]
    if base_type in generator.__bases__:
        return generator
    return type(
        generator.__name__,
        tuple(list(generator.__bases__) + [base_type]),
        generator.__dict__ | base_type.__dict__,
    )
